import os
from commonfate_provider.config import loaders


def test_env_loader_works():
    loader = loaders.EnvLoader()
    os.environ["PROVIDER_CONFIG_MY_VALUE"] = "something"
    got = loader.load_string("my_value")

    assert got == "something"


def test_dev_env_secret_loader_works():
    loader = loaders.DevEnvSecretLoader()
    os.environ["PROVIDER_SECRET_MY_VALUE"] = "something"
    got = loader.load_secret_string("my_value")

    assert got == loaders.Secret(
        ref="env://PROVIDER_SECRET_MY_VALUE", value="something"
    )


def test_dict_loader_works():
    loader = loaders.DictLoader({"value": "something"})
    got = loader.load_string("value")

    assert got == "something"


def test_dict_loader_secrets_works():
    loader = loaders.DictLoader({"value": "something"})
    got = loader.load_secret_string("value")

    want = loaders.Secret(ref="dict://value", value="something")

    assert got == want



from dataclasses import dataclass
from commonfate_provider import (
    provider,
    target,
    resources,
    tasks,
    namespace,
    schema,
)
import typing

from pydantic import BaseModel, Field


class GrantData(BaseModel):
    class Target(BaseModel):
        arguments: typing.Dict[str, str]
        kind: str

    subject: str
    target: Target


class Grant(BaseModel):
    type: typing.Literal["grant"]
    data: GrantData


class Revoke(BaseModel):
    type: typing.Literal["revoke"]
    data: GrantData


class Describe(BaseModel):
    type: typing.Literal["describe"]


class Load(BaseModel):
    class Data(BaseModel):
        task: str
        """the resource loader function ID to run"""
        ctx: typing.Optional[dict] = {}
        """context information for the task"""

    type: typing.Literal["load"]
    data: Data


class Event(BaseModel):
    __root__: typing.Union[Grant, Revoke, Load, Describe] = Field(
        ..., discriminator="type"
    )


_T = typing.TypeVar("_T")


class DescribeResponse(BaseModel):
    provider: dict
    config: dict
    diagnostics: typing.List[dict]
    healthy: bool
    provider_schema: dict = Field(alias="schema")


class LoadResponse(BaseModel):
    resources: typing.List[dict]
    tasks: typing.List[dict]


class Result(BaseModel):
    response: typing.Union[DescribeResponse, LoadResponse]


@dataclass
class AWSLambdaRuntime:
    provider: provider.Provider
    name: typing.Optional[str] = None
    version: typing.Optional[str] = None
    publisher: typing.Optional[str] = None

    def handle(self, event, context):
        result = self._do_handle(event=event, context=context)
        if result is not None:
            return result.dict(by_alias=True)

    def _do_handle(self, event, context) -> typing.Optional[Result]:
        parsed = Event.parse_obj(event)
        event = parsed.__root__

        if isinstance(event, Grant):
            try:
                registered_targets = namespace.get_target_classes()
                registered_target = registered_targets[event.data.target.kind]
                args_cls = registered_target.cls
            except KeyError:
                all_keys = ",".join(registered_targets.keys())
                raise KeyError(
                    f"unhandled target kind {event.data.target.kind}, supported kinds are [{all_keys}]"
                )

            args = target._initialise(args_cls, event.data.target.arguments)
            grant = registered_target.get_grant_func()
            grant(self.provider, event.data.subject, args)
            return None

        elif isinstance(event, Revoke):
            try:
                registered_targets = namespace.get_target_classes()
                registered_target = registered_targets[event.data.target.kind]
                args_cls = registered_target.cls
            except KeyError:
                all_keys = ",".join(registered_targets.keys())
                raise KeyError(
                    f"unhandled target kind {event.data.target.kind}, supported kinds are [{all_keys}]"
                )

            args = target._initialise(args_cls, event.data.target.arguments)

            revoke = registered_target.get_revoke_func()
            revoke(self.provider, event.data.subject, args)
            return None

        if isinstance(event, Describe):
            # Describe returns the configuration of the provider including the current status.
            provider = {
                "publisher": self.publisher,
                "name": self.name,
                "version": self.version,
            }
            config = self.provider._safe_config
            diagnostics = self.provider.diagnostics.export_logs()
            healthy = self.provider.diagnostics.has_no_errors()
            provider_schema = schema.export_schema().dict(exclude_none=True)

            response = DescribeResponse(
                config=config,
                diagnostics=diagnostics,
                healthy=healthy,
                provider=provider,
                schema=provider_schema,
            )

            return Result(response=response)

        elif isinstance(event, Load):
            resources._reset()
            tasks._reset()
            tasks._execute(
                provider=self.provider, task=event.data.task, ctx=event.data.ctx
            )
            # find the resources and pending tasks, and return them
            found = resources.get()
            pending_tasks = tasks.get()

            response = LoadResponse(
                resources=[r.export_json() for r in found],
                tasks=[t.json() for t in pending_tasks],
            )

            return Result(response=response)

        else:
            raise Exception(f"unhandled event type")



from dataclasses import dataclass, asdict
import typing


@dataclass
class Log:
    level: typing.Literal["INFO", "ERROR"]
    msg: str


class Logs:
    logs: typing.List[Log]

    def __init__(self) -> None:
        self.logs = []

    def info(self, msg: str):
        self.logs.append(Log("INFO", msg=msg))

    def error(self, msg: str):
        self.logs.append(Log("ERROR", msg=msg))

    def has_no_errors(self) -> bool:
        has_error = any([l for l in self.logs if l.level == "ERROR"])
        return not has_error

    def export_logs(self) -> typing.List[dict]:
        """
        Returns the diagnostic logs as a list of simple dict objects
        so that it can be easily serialised as JSON.
        """
        return [asdict(l) for l in self.logs]



from commonfate_provider import provider
from commonfate_provider.runtime import initialise


def test_lambda_runtime_calls_provider_setup():
    class MyProvider(provider.Provider):
        def setup(self):
            self.is_setup = True

    p = initialise.initialise_provider()
    assert p.is_setup == True



"""
Module namespace as an internal module which contains logic to manage the registered
Provider and decorated functions, like Targets and grant/revoke
methods.

For testing purposes, namespace.clear() can be used to remove
all entries from the namespace between different tests.
"""

from dataclasses import dataclass
import inspect
import typing

if typing.TYPE_CHECKING:
    from commonfate_provider.provider import Provider, ConfigValidator
    from commonfate_provider.resources import Resource
    from commonfate_provider.access import (
        GrantFunc,
        RevokeFunc,
        GrantValidator,
    )


_PROVIDER: typing.Optional[typing.Type["Provider"]] = None


_ALL_RESOURCES: typing.List["Resource"] = []
"""instances of particular resources. Updated when resources.register() is called."""

_RESOURCE_CLASSES: typing.List[typing.Type["Resource"]] = []
"""the resource classes themselves. Updated when a class subclasses resources.Resource"""

_ALL_GRANT_VALIDATORS: typing.Dict[str, typing.Dict[str, "GrantValidator"]] = {}

_ALL_CONFIG_VALIDATORS: typing.Dict[str, "ConfigValidator"] = {}


@dataclass
class Target:
    cls: typing.Any
    grant_func: typing.Optional["GrantFunc"] = None
    revoke_func: typing.Optional["RevokeFunc"] = None

    def get_revoke_func(self) -> "RevokeFunc":
        if self.revoke_func is None:
            raise Exception("revoke function is not defined for this target")
        return self.revoke_func

    def get_grant_func(self) -> "GrantFunc":
        if self.grant_func is None:
            raise Exception("grant function is not defined for this target")
        return self.grant_func


_TARGET_CLASSES: typing.Dict[str, Target] = {}


LoaderFunc = typing.Callable[[typing.Any], None]


_RESOURCE_LOADERS: typing.Dict[str, LoaderFunc] = {}


def register_provider(provider_class: typing.Type["Provider"]):
    """
    Register a provider class.
    """
    global _PROVIDER
    if _PROVIDER is not None:
        raise Exception(
            f"Tried to register a Provider class {provider_class.__name__} but an existing class has already been registered ({_PROVIDER.__name__})"
        )

    _PROVIDER = provider_class


def register_target_class(kind: str, target_class: typing.Type[typing.Any]):
    _TARGET_CLASSES[kind] = Target(cls=target_class)


_T = typing.TypeVar("_T")

KindType = typing.Optional[typing.Union[typing.Type[_T], str]]


def register_grant_func(kind: KindType, func: "GrantFunc"):
    kind_str = _lookup_kind(kind=kind, function_type="grant function")
    _TARGET_CLASSES[kind_str].grant_func = func


def register_revoke_func(kind: KindType, func: "RevokeFunc"):
    kind_str = _lookup_kind(kind=kind, function_type="revoke function")
    _TARGET_CLASSES[kind_str].revoke_func = func


def _lookup_kind(kind: KindType, function_type: str) -> str:
    """
    Looks up the `kind` parameter to match it with a target kind.

    The `kind` parameter is optional in our decorator methods,
    so if it isn't provided, we try and look up the registered target class.
    """
    if len(_TARGET_CLASSES) == 0:
        raise Exception(
            f"A target class must be defined with @access.target() before registering a {function_type}"
        )
    if kind is None and len(_TARGET_CLASSES) > 1:
        raise Exception(
            f'This provider grants access to multiple targets. You must specify the kind of target that this {function_type} refers to. For example: @decorator_method(kind="MyTarget")'
        )

    if kind is None:
        kind = next(iter(_TARGET_CLASSES))

    if inspect.isclass(kind):
        kind = kind.__name__

    return kind


def register_resource_class(resource_class: typing.Type["Resource"]):
    _RESOURCE_CLASSES.append(resource_class)


def register_resource_loader(func: LoaderFunc):
    _RESOURCE_LOADERS[func.__name__] = func


def register_grant_validator(
    kind: KindType, id: str, grant_validator: "GrantValidator"
):
    kind_str = _lookup_kind(kind=kind, function_type="grant validator")
    _ALL_GRANT_VALIDATORS.setdefault(kind_str, {})
    _ALL_GRANT_VALIDATORS[kind_str][id] = grant_validator


def register_config_validator(id: str, config_validator: "ConfigValidator"):
    _ALL_CONFIG_VALIDATORS[id] = config_validator


def get_provider() -> typing.Type["Provider"]:
    """
    Loads the registered Provider class.
    """
    global _PROVIDER
    if _PROVIDER is None:
        raise Exception(
            "a Provider has not been registered. Make sure that you add a provider class which subclasses provider.Provider."
        )
    return _PROVIDER


def get_resource_classes() -> typing.List[typing.Type["Resource"]]:
    return _RESOURCE_CLASSES


def get_resource_loaders() -> typing.Dict[str, LoaderFunc]:
    return _RESOURCE_LOADERS


def get_target_classes() -> typing.Dict[str, typing.Type[Target]]:
    return _TARGET_CLASSES


def get_config_validators():
    return _ALL_CONFIG_VALIDATORS


def clear():
    """
    Removes all registered classes and functions.

    Used for testing only.
    """
    global _PROVIDER, _ALL_RESOURCES, _RESOURCE_CLASSES, _RESOURCE_LOADERS, _TARGET_CLASSES, _ALL_GRANT_VALIDATORS
    _PROVIDER = None
    _ALL_RESOURCES = []
    _RESOURCE_CLASSES = []
    _RESOURCE_LOADERS = {}
    _TARGET_CLASSES = {}
    _ALL_GRANT_VALIDATORS = {}



import pytest
from syrupy.extensions.json import JSONSnapshotExtension

from commonfate_provider import access, namespace, resources, target


@pytest.fixture(autouse=True)
def fresh_namespace():
    """clear the registered provider, targets etc between test runs"""
    yield
    namespace.clear()


@pytest.fixture
def snapshot_json(snapshot):
    """use JSON, rather than AmberSnapshotExtension as our schema is serialized as JSON"""
    return snapshot.use_extension(JSONSnapshotExtension)


@access.target()
class ExampleArgs:
    group = target.String()


def test_parse_args():
    got_target = target._initialise(ExampleArgs, {"group": "test"})
    assert got_target.group == "test"


def test_parse_args_missing_required():
    with pytest.raises(target.ParseError):
        target._initialise(ExampleArgs, {})


def test_export_target_schema(snapshot_json):
    @access.target()
    class ExampleTarget:
        my_property = target.String(title="MyProperty")

    got = target.export_schema()
    got_dict = {k: v.dict() for k, v in got.items()}

    assert got_dict == snapshot_json


def test_export_target_resource_schema(snapshot_json):
    class MyResource(resources.Resource):
        pass

    @access.target()
    class ExampleTarget:
        my_property = target.String(title="MyProperty", description="some description")
        my_resource = target.Resource(title="MyResourceField", resource=MyResource)

    got = target.export_schema()
    got_dict = {k: v.dict() for k, v in got.items()}

    assert got_dict == snapshot_json



import typing
from pkg_resources import get_distribution
from commonfate_provider import namespace, resources, target
from common_fate_schema.provider import v1alpha1


def export_schema() -> v1alpha1.Schema:
    """
    Export a Provider schema, ready to be serialized in JSON format.
    """

    Provider = namespace.get_provider()

    framework_version = None
    try:
        framework_version = get_distribution("commonfate_provider").version
    except Exception:
        pass

    config_schema = Provider.export_config_schema()
    resources_schema = resources.export_schema()

    targets = target.export_schema()

    schema = v1alpha1.Schema(
        targets=targets,
        config=config_schema,
        resources=resources_schema,
        meta=v1alpha1.Meta(framework=framework_version),
    )

    return schema



import json
import pytest
from syrupy.extensions.json import JSONSnapshotExtension

from commonfate_provider import access, namespace, provider, resources, schema, target


@pytest.fixture(autouse=True)
def fresh_namespace():
    """clear the registered provider, targets etc between test runs"""
    yield
    namespace.clear()


@pytest.fixture
def snapshot_json(snapshot):
    """use JSON, rather than AmberSnapshotExtension as our schema is serialized as JSON"""
    return snapshot.use_extension(JSONSnapshotExtension)


def test_simple_schema(snapshot_json):
    class Provider(provider.Provider):
        pass

    @access.target()
    class MyTarget:
        pass

    actual = schema.export_schema()
    assert actual == snapshot_json


def test_schema_with_config(snapshot_json):
    class ConfigResource(resources.Resource):
        some_val: str

    class Provider(provider.Provider):
        some_config = provider.String(description="Some config usage")
        api_token = provider.String(description="API Token", secret=True)

    @access.target()
    class MyTarget:
        first = target.String(description="first var", title="First")
        my_resource = target.Resource(
            description="first var", title="First", resource=ConfigResource
        )

    actual = schema.export_schema()
    assert actual == snapshot_json


def test_multiple_targets(snapshot_json):
    class Provider(provider.Provider):
        pass

    @access.target()
    class MyTarget:
        first = target.String(description="first var", title="First")

    @access.target()
    class SecondTarget:
        second = target.String(description="second var", title="Second")

    actual = schema.export_schema()
    assert actual == snapshot_json



from dataclasses import dataclass
import typing
from commonfate_provider import namespace, resources
from common_fate_schema.provider import v1alpha1


class ParseError(Exception):
    """Raised if Provider arguments cannot be parsed"""


@dataclass
class Option:
    value: str
    label: str
    description: typing.Optional[str] = None


_T = typing.TypeVar("_T")


def _initialise(cls: type[_T], raw_targets: dict) -> _T:
    # initialise an instance of the Target class 'cls'
    instance = cls()

    # assign each variable in the class based on the 'raw_targets' dict
    all_vars = [k for k in vars(cls).keys() if not k.startswith("__")]
    for k in all_vars:
        if k not in raw_targets:
            raise ParseError(f"{k} argument is required")
        val = raw_targets[k]
        setattr(instance, k, val)

    return instance


@dataclass
class Field:
    title: typing.Optional[str] = None
    description: typing.Optional[str] = None


class String(Field):
    pass


@dataclass
class Resource(Field):
    resource: typing.Optional[resources.Resource] = None


def export_schema() -> typing.Dict[str, v1alpha1.Target]:
    """
    Exports the target schema defined in an Target class
    to a dictionary.

    For example, if Target is defined as:
    ```
    class Target(target.Target):
        group = target.String(title="Group", description="group to grant access to")
    ```

    this method will produce a dictionary in the form:
    ```json
    {
        "group": {
            "id": "group",
            "description": "group to grant access to",
            "title": "Group",
            "type": "string"
        }
    }
    ```
    """
    all_targets: typing.Dict[str, v1alpha1.Target] = {}

    for kind, registered_target in namespace.get_target_classes().items():
        target_class = registered_target.cls
        target_schema: typing.Dict[str, v1alpha1.TargetField] = {}
        all_vars = [
            (k, v) for (k, v) in vars(target_class).items() if not k.startswith("__")
        ]
        for k, v in all_vars:
            if type(v) == String:
                val: String = v
                schema = v1alpha1.TargetField(
                    title=val.title,
                    type="string",
                    resource=None,
                    description=val.description,
                )

                target_schema[k] = schema
            if type(v) == Resource:
                val: Resource = v
                schema = {
                    "id": k,
                    "type": "string",
                    "title": val.title,
                    "resourceName": val.resource.__name__,
                }

                schema = v1alpha1.TargetField(
                    title=val.title,
                    type="string",
                    resource=val.resource.__name__,
                    description=val.description,
                )

                target_schema[k] = schema

        all_targets[kind] = v1alpha1.Target(type="object", properties=target_schema)

    return all_targets



import typing

from commonfate_provider import config, provider
from commonfate_provider.config import loaders
from commonfate_provider.runtime import initialise


def initialise_test_provider(config_dict: typing.Dict[str, str]) -> provider.Provider:
    """
    A wrapper around commonfate_provider.runtime.initialise
    which uses a dictionary for all config values.

    For use in tests only.
    """
    loader = loaders.DictLoader(config_dict=config_dict)
    configurer = config.Configurer(string_loader=loader, secret_string_loader=loader)

    return initialise.initialise_provider(configurer=configurer)



import pytest
from commonfate_provider import access, namespace


@pytest.fixture(autouse=True)
def fresh_namespace():
    """clear the registered provider, targets etc between test runs"""
    yield
    namespace.clear()


def test_target():
    @access.target()
    class ExampleTarget:
        pass

    targets = namespace.get_target_classes()

    assert targets["ExampleTarget"].cls == ExampleTarget


def test_target_with_kind():
    @access.target(kind="SomethingElse")
    class ExampleTarget:
        pass

    targets = namespace.get_target_classes()

    assert targets["SomethingElse"].cls == ExampleTarget



import typing
from commonfate_provider import config, namespace, provider


def initialise_provider(
    configurer: typing.Optional[config.Configurer] = None,
) -> provider.Provider:
    """
    Used in the Provider runtimes to load and initialise a provider class.
    """
    # load the provider from the provider.Provider subclass
    Provider = namespace.get_provider()

    # initialise the provider class
    provider = Provider()

    # configure the provider
    if configurer is not None:
        configurer.configure(provider)

    # run the setup method to call any provider developer-defined
    # setup operations, like configuring API clients.
    provider.setup()

    return provider



from dataclasses import dataclass
import typing_extensions
import typing

from commonfate_provider import namespace


_T = typing.TypeVar("_T")


@typing_extensions.dataclass_transform()
def target(
    kind: typing.Optional[str] = None,
) -> typing.Callable[[type[_T]], type[_T]]:
    """
    Define a target for access.

    Example:
    ```
    @access.target()
    class Target:
        target_property = target.String()
    ```
    """

    def actual_decorator(cls: type[_T]) -> type[_T]:
        nonlocal kind
        if kind is None:
            kind = cls.__name__

        namespace.register_target_class(kind, cls)

        return cls

    return actual_decorator


@dataclass
class GrantResult:
    access_instructions: typing.Optional[str] = None
    """
    Instructions on how to access the entitlements.
    Common Fate will display these to the user upon a successful Access Request.
    """


GrantValidatorFunc = typing.Callable[[typing.Any, str, typing.Any], None]


@dataclass
class GrantValidator:
    name: str
    func: GrantValidatorFunc


def grant_validator(
    name: str,
    kind: namespace.KindType = None,
) -> typing.Callable[[GrantValidatorFunc], GrantValidatorFunc]:
    def actual_decorator(func: GrantValidatorFunc):
        id = func.__name__
        gv = GrantValidator(id=id, name=name, func=func)
        namespace.register_grant_validator(kind=kind, grant_validator=gv)
        return func

    return actual_decorator


GrantFunc = typing.Callable[[typing.Any, str, typing.Any], typing.Optional[GrantResult]]

_T = typing.TypeVar("_T")


def grant(
    kind: namespace.KindType = None,
) -> typing.Callable[[GrantFunc], GrantFunc]:
    """
    Register a function as an access granter.

    For example:
    ```
    @access.grant()
    def grant(p: Provider, subject: str, target: MyTargetClass):
        ...
    ```

    The `kind` parameter may be specified to indicate a particular target kind
    that this function grants access to.

    For example:
    ```
    @access.target(kind="MyTargetKind")
    class Target:
        ...

    @access.grant(kind="MyTargetKind")
    def grant(p: Provider, subject: str, target: MyTargetClass):
        ...
    ```
    """

    def actual_decorator(func: GrantFunc):
        namespace.register_grant_func(kind=kind, func=func)
        return func

    return actual_decorator


RevokeFunc = typing.Callable[[typing.Any, str, typing.Any], None]


def revoke(
    kind: typing.Optional[typing.Union[typing.Type[_T], str]] = None,
) -> typing.Callable[[RevokeFunc], RevokeFunc]:
    """
    Register a function as an access revoker.

    For example:
    ```
    @access.revoke()
    def revoke(p: Provider, subject: str, target: MyTargetClass):
        ...
    ```

    The `kind` parameter may be specified to indicate a particular target kind
    that this function revokes access to.

    For example:
    ```
    @access.target(kind="MyTargetKind")
    class Target:
        ...

    @access.revoke(kind="MyTargetKind")
    def revoke(p: Provider, subject: str, target: MyTargetClass):
        ...
    ```
    """

    def actual_decorator(func: RevokeFunc):
        namespace.register_revoke_func(kind=kind, func=func)
        return func

    return actual_decorator
